//
//  tank
//
//  Import and Playlist Daemon for autoradio project
//
//
//  Copyright (C) 2017-2018 Christian Pointner <equinox@helsinki.at>
//
//  This file is part of tank.
//
//  tank is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  any later version.
//
//  tank is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with tank. If not, see <http://www.gnu.org/licenses/>.
//

package importer

import (
	"fmt"
	"io"
	"os"
	"os/exec"
	"strconv"
	"syscall"
)

type converter interface {
	io.WriteCloser
	Wait() (loudnessCorr float64, log []string, err error)
}

func (job *Job) newConverter() (converter, error) {
	switch job.im.conf.Converter {
	case ConvNull:
		return newNullConverter(job)
	case ConvFFmpeg:
		return newFFmpegConverter(job)
	}
	panic("invalid fetch converter")
}

//******* null

type nullConverter struct {
	job  *Job
	file *os.File
	log  []string
}

func newNullConverter(job *Job) (c *nullConverter, err error) {
	c = &nullConverter{job: job}
	filename := job.im.store.GetFilePath(job.Group, job.ID)
	job.im.dbgLog.Printf("null-converter: opened file '%s'", filename)
	if c.file, err = os.OpenFile(filename, os.O_WRONLY|os.O_CREATE|os.O_EXCL, 0600); err != nil {
		c.log = append(c.log, "ERROR opening file failed: "+err.Error())
	} else {
		c.log = append(c.log, "successfully opened file: "+filename)
	}
	return
}

func (c *nullConverter) Write(p []byte) (n int, err error) {
	return c.file.Write(p)
}

func (c *nullConverter) Close() (err error) {
	filename := c.file.Name()
	if err = c.file.Close(); err != nil {
		c.log = append(c.log, "ERROR closing file failed: "+err.Error())
	} else {
		c.log = append(c.log, "successfully closed file: "+filename)
	}
	return
}

func (c *nullConverter) Wait() (loudnessCorr float64, log []string, err error) {
	return 0, c.log, nil
}

//******* FFmpeg

type ffmpegConverter struct {
	job    *Job
	log    *convLog
	cmd    *exec.Cmd
	stdin  io.WriteCloser
	stdout *convLogger
	stderr *convLogger
}

func newFFmpegConverter(job *Job) (c *ffmpegConverter, err error) {
	c = &ffmpegConverter{job: job}
	c.log = newConvLog()
	filename := job.im.store.GetFilePath(job.Group, job.ID)
	job.im.dbgLog.Printf("ffmpeg-converter: starting ffmpeg for file '%s'", filename)

	// c.cmd = exec.CommandContext(job.ctx, "ffmpeg", "-hide_banner", "-nostats", "-i", "-")
	c.cmd = exec.CommandContext(job.ctx, "ffmpeg", "-nostats", "-y", "-i", "-")
	c.cmd.Args = append(c.cmd.Args, "-map_metadata", "0", "-vn")
	c.cmd.Args = append(c.cmd.Args, "-ar", strconv.FormatUint(uint64(job.im.store.Audio.SampleRate), 10))
	c.cmd.Args = append(c.cmd.Args, "-f", job.im.store.Audio.Format.String(), filename)
	// loudness normalization, see: http://k.ylo.ph/2016/04/04/loudnorm.html
	c.cmd.Args = append(c.cmd.Args, "-map_metadata", "-1", "-vn")
	c.cmd.Args = append(c.cmd.Args, "-ar", "192k")
	c.cmd.Args = append(c.cmd.Args, "-filter:a", "loudnorm=print_format=json:dual_mono=true")
	c.cmd.Args = append(c.cmd.Args, "-f", "null", "/dev/null")

	if c.stdin, err = c.cmd.StdinPipe(); err != nil {
		c.log.Append("ERROR opening stdin pipe: " + err.Error())
		return nil, err
	}

	var stdout, stderr io.Reader
	if stdout, err = c.cmd.StdoutPipe(); err != nil {
		c.log.Append("ERROR opening stdout pipe: " + err.Error())
		return nil, err
	}
	c.stdout = newConvLogger(c.log, "ffmpeg(stderr)> ", stdout)
	if stderr, err = c.cmd.StderrPipe(); err != nil {
		c.log.Append("ERROR opening stderr pipe: " + err.Error())
		return nil, err
	}
	c.stderr = newConvLogger(c.log, "ffmpeg(stderr)> ", stderr)

	if err = c.cmd.Start(); err != nil {
		c.log.Append("ERROR starting ffmpeg: " + err.Error())
		return nil, err
	}
	return
}

func (c *ffmpegConverter) Write(p []byte) (n int, err error) {
	return c.stdin.Write(p)
}

func (c *ffmpegConverter) Close() (err error) {
	return c.stdin.Close()
}

func (c *ffmpegConverter) Wait() (loudnessCorr float64, log []string, err error) {
	c.stdout.Wait()
	c.stderr.Wait()
	err = c.cmd.Wait()

	if err != nil {
		if exiterr, ok := err.(*exec.ExitError); ok {
			if status, ok := exiterr.Sys().(syscall.WaitStatus); ok {
				exitcode := status.ExitStatus()
				c.job.im.dbgLog.Printf("ffmpeg-converter: ffmpeg returned %d", exitcode)
				c.log.Append(fmt.Sprintf("ffmpeg returned %d", exitcode))
			} else {
				c.log.Append("ERROR getting exit code of ffmpeg: " + err.Error())
				c.job.im.errLog.Println("ffmpeg-converter: getting exit code of ffmpeg failed:", err)
			}
		} else {
			c.log.Append("ERROR getting exit code of ffmpeg: " + err.Error())
			c.job.im.errLog.Println("ffmpeg-converter: getting exit code of ffmpeg failed:", err)
		}
	}
	return 0, c.log.log, err
}
